import json
import os
import os.path
from typing import Any, Dict, Iterable, List, Optional

from cerberus import Validator

FILE_SPEC_ITEM_SCHEMA = {
    'file': {'type': 'string', 'required': True},
    'metadata': {
        'type': 'dict',
        'keyschema': {'type': 'string'},
        'valueschema': {'nullable': False},
        'required': False
    }
}


class InvalidPathException(Exception):

    def __init__(self, path):
        self.path = path


class FileSpecFormatError(Exception):

    def __init__(self, path, message):
        self.path = path
        self.message = message


class UploadFile:

    def __init__(self, path, metadata=None):
        self.path = path
        self.metadata = metadata


class UploadFileSpec:

    def __init__(self):
        self.paths = set()
        self.__metadata_by_path = {}

    def add_metadata(self, path: str, metadata: Dict[str, Any]):
        full_path = os.path.realpath(path)
        self.__metadata_by_path[full_path] = metadata

    def get_metadata(self, path: str) -> Optional[Dict[str, Any]]:
        full_path = os.path.realpath(path)
        return self.__metadata_by_path.get(full_path)

    @staticmethod
    def parse(file_list_path: str) -> 'UploadFileSpec':
        """Parses a spec file located at ``path``.

        Args:
            path: Spec file path. Spec file can be JSON.

        Returns:
            A ``UploadFileSpec`` object.

        Raises:
            FileNotFoundError: The file doesn't exist at ``file_list_path``.
            FileSpecFormatError: Validation error occurred.
        """
        try:
            with open(file_list_path, 'r', encoding='utf-8') as f:
                specs = json.load(f)
        except json.JSONDecodeError as error:
            raise FileSpecFormatError(
                file_list_path,
                'The file {} was malformed'.format(file_list_path)) from error

        specs_obj = UploadFileSpec()

        if not isinstance(specs, list):
            raise FileSpecFormatError(file_list_path, 'The top level must be an array.')

        for spec in specs:
            validator = Validator(FILE_SPEC_ITEM_SCHEMA)
            if not validator.validate(spec):
                raise FileSpecFormatError(
                    file_list_path,
                    'File spec item validation failed: {}.'.format(validator.errors))

            filepath = spec['file']
            specs_obj.paths.add(filepath)
            metadata = spec.get('metadata')
            if metadata:
                specs_obj.add_metadata(filepath, metadata)

        return specs_obj


def generate_upload_file_iter(
        paths: List[str],
        file_list_path: Optional[str]=None,
        recursive: bool=False) -> Iterable[UploadFile]:
    """Returns a generator which yields ``UploadFile`` for specified paths and directories.

    Args:
        paths: Uploading file paths.
        file_list_path: JSON file specifies the list of upload files.
        recursive: ``True`` if this function search files under the directories.

    Returns:
        A generator which yields ``UploadFile``.
    """
    spec = None

    if file_list_path:
        spec = UploadFileSpec.parse(file_list_path)

    if not spec:
        spec = UploadFileSpec()

    spec.paths.update(paths)

    for path in spec.paths:
        if os.path.isfile(path):
            yield UploadFile(path, spec.get_metadata(path))
        elif os.path.isdir(path):
            if not recursive:
                raise InvalidPathException(path)
            for root, _, file_paths in os.walk(path):
                for file_path in file_paths:
                    full_path = os.path.join(root, file_path)
                    yield UploadFile(full_path, spec.get_metadata(full_path))
        else:
            raise InvalidPathException(path)
