#!/usr/bin/env python3
# -*- coding: utf-8 -*

import datetime
import json
import os
import sys
import tarfile
import time
import mimetypes
import tempfile
import base64
from concurrent.futures import ThreadPoolExecutor

import click
import requests

from json import JSONDecodeError

from pygments import highlight, lexers, formatters
from pygments.styles import get_style_by_name
from six.moves import input

import abejacli.version
from abejacli.config import (
    WEB_API_ENDPOINT,
    ROOT_DIRECTORY,
    ORGANIZATION_ENDPOINT,
    ABEJA_API_URL,
    CONFIG_FILE_PATH,
    SAMPLE_MODEL_PATH,
    SUCCESS_EXITCODE,
    ERROR_EXITCODE,
    CONFIG,
    TRIGGER_DEFAULT_RETRY_COUNT,
    RUN_DEFAULT_RETRY_COUNT
)
from abejacli.session import (
    api_get,
    api_post,
    api_delete,
    api_patch,
    download_to_file,
    api_get_data,
)
from abejacli.fs_utils import (
    generate_upload_file_iter,
    InvalidPathException
)
from abejacli.datalake import (
    upload_to_datalake,
    download_from_datalake,
    generate_channel_file_iter_by_id,
    generate_channel_file_iter_by_period,
)
from abejacli.dataset import import_dataset_from_datalake
from abejacli.model.docker_handler import LocalModelHandler
from abejacli.model.local_server_manager import LocalServerManager
from abejacli.click_custom import (
    MutuallyExclusiveAndRequireOption,
    DATE_STR,
    METADATA_STR,
    ENVIRONMENT_STR,
    USER_PARAM_STR,
    DATASET_PARAM_STR
)

from abejacli.training import TrainingConfig
from abejacli.exceptions import InvalidConfigException, ConfigFileNotFoundError
from abejacli.logger import get_logger

__version__ = abejacli.version.VERSION
date = datetime.datetime.today()
logger = get_logger()
training_config = TrainingConfig()


def json_output_formatter(parsed_json):
    """
    Applies syntax highlighting to output
    :params parsed_json: (dict) JSON response requests response.json()
    :returns: (str) syntax highlighted text
    """
    formatted_json = json.dumps(parsed_json,
                                sort_keys=True,
                                ensure_ascii=False,
                                indent=4)
    output_style = get_style_by_name('emacs')
    colorful_json = highlight(formatted_json,
                              lexers.JsonLexer(),
                              formatters.Terminal256Formatter(style=output_style))
    return colorful_json


@click.group(help='ABEJA command line interface')
@click.version_option(version=__version__, message='v%(version)s')
@click.pass_context
def main(ctx):
    pass


@main.group(help='Model operation commands')
@click.pass_context
def model(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@main.group(help='DataLake operation commands')
@click.pass_context
def datalake(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@main.group(help='Schema operation commands')
@click.pass_context
def schema(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@main.group(help='Database operation commands')
@click.pass_context
def database(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@main.group(help='Training operation commands')
@click.pass_context
def training(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


@main.group(help='Dataset operation commands')
@click.pass_context
def dataset(ctx):
    if not CONFIG:
        click.echo("[error] there is no configuration, execute 'abeja configure'")
        ctx.abort()
        return


# ---------------------------------------------------
# Configure
# ---------------------------------------------------


@main.command(help='Set ABEJA credentials')
@click.pass_context
def configure(ctx):
    user = input('abeja-platform-user  : ').strip()
    token = input('personal-access-token: ').strip()
    org = input('organization-name    : ').strip()

    config = {
        "abeja-platform-user": "user-{}".format(user),
        "personal-access-token": token,
        "organization-name": org
    }

    if not all(config.values()):
        click.echo("[error] any of user-id, access-token, organization-name is not set")
        ctx.abort()
        return

    _create_configfile(config)
    click.echo('[INFO]: ABEJA credentials setup completed!')


def _create_configfile(config):
    if not os.path.exists(ROOT_DIRECTORY):
        os.makedirs(ROOT_DIRECTORY, mode=0o711)

    with open(CONFIG_FILE_PATH, "w") as f:
        json.dump(config, f)


# ---------------------------------------------------
# models
# ---------------------------------------------------
@model.command(name='create-model', help='Create model')
@click.option('-n', '--name', 'name', type=str, help='Model name', required=True)
@click.option('-d', '--description', 'description', type=str, help='Model description', required=True)
@click.option('-i', '--init_project', 'init_project', is_flag=True, help='Set if want to init project')
def create_model(name, description, init_project):
    try:
        r = _create_model(name, description, init_project)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _init_project(name):
    if not os.path.exists(name):
        os.makedirs(name)
    download_to_file(name, "{}/cli/main.py".format(SAMPLE_MODEL_PATH))


def _create_model(name, description, init_project=False):
    parameter = {
        'name': name,
        'description': description,
    }
    json_data = json.dumps(parameter)

    # create directory & sample code download
    if init_project:
        _init_project(name)

    url = "{}/models".format(ORGANIZATION_ENDPOINT)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-models', help='Get to model or endpoint model')
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', default='all', required=False)
def describe_models(model_id):
    try:
        r = _describe_models(model_id)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _describe_models(model_id):
    url = '{}/models'.format(ORGANIZATION_ENDPOINT)
    if model_id != "all":
        url = '{}/{}'.format(url, model_id)
    r = api_get(url)

    return r


@model.command(name='delete-model', help='Delete model')
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', required=True)
def delete_model(model_id):
    try:
        r = _delete_model(model_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_model(model_id):
    url = '{}/models/{}'.format(ORGANIZATION_ENDPOINT, model_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# models versions
# ---------------------------------------------------
@model.command(name='create-version', help='Create version & upload application')
@click.pass_context
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', required=True)
@click.option('-v', '--version', 'version', type=str, help='Model version code', required=True)
@click.option('-i', '--image', 'image', type=str, help='Base-image name', required=True)
@click.option('-h', '--handler', 'handler', type=str, help='Path to handler in the model archive.')
@click.option('-j', '--job_definition_name', 'job_definition_name', type=str, help='Training job definition name',
              required=False)
@click.option('-t', '--training_job_id', 'training_job_id', type=str, help='Training Job id', required=False)
def create_version(ctx, model_id, version, image, handler, job_definition_name, training_job_id):
    try:
        r = _create_version(ctx, model_id, version, image, handler, job_definition_name, training_job_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _create_version(ctx, model_id, version, image, handler=None, job_definition_name=None, training_job_id=None):
    parameter = {
        'image': image,
        'version': version
    }
    if handler:
        parameter['handler'] = handler
    if job_definition_name:
        parameter['job_definition_name'] = job_definition_name
    if training_job_id:
        parameter['training_job_id'] = training_job_id
    json_data = json.dumps(parameter)

    url = '{}/models/{}/versions'.format(ORGANIZATION_ENDPOINT, model_id)
    r = api_post(url, json_data)
    result = {
        'version': r['version'],
        'version_id': r['version_id']
    }

    upload_url = r['upload_url']
    name = _get_model_name(model_id)
    try:
        temp_file = _version_archive(name)
        _version_upload(upload_url, temp_file.name)  # TODO:  Confirm what is this doing?  Should we use it's result?
    finally:
        if os.path.exists(temp_file.name):
            temp_file.close()
            os.unlink(temp_file.name)
    return result


def _get_model_name(model_id):
    url = '{}/models/{}'.format(ORGANIZATION_ENDPOINT, model_id)
    r = api_get(url)
    return r['name']


def _version_archive(name):
    """
    Create .tar.gz archive file for current directory excluding .git
    Set model name as archive name prefix
    :param: name
    :return: temporary archive file
    """

    filenames = os.listdir(os.getcwd())
    execlude_filenames = ['.git']
    tmp_file = tempfile.NamedTemporaryFile(prefix=name, suffix='.tar.gz', delete=False)
    with tarfile.open(fileobj=tmp_file, mode="w:gz") as tar:
        for filename in filenames:
            if filename in execlude_filenames:
                continue
            tar.add(filename)
    tmp_file.seek(0)
    return tmp_file


def _version_upload(upload_url, tar_name):
    with open(tar_name, "rb") as file:
        result = requests.put(upload_url, data=file)

    return result


@model.command(name='describe-versions', help='Get to version or version list')
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', required=True)
@click.option('-v', '--version_id', 'version_id', type=str, help='Version identifier', default='all', required=False)
def describe_versions(model_id, version_id):
    try:
        r = _describe_versions(model_id, version_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_versions(model_id, version_id):
    url = '{}/models/{}/versions'.format(ORGANIZATION_ENDPOINT, model_id)
    if version_id != "all":
        url = '{}/{}'.format(url, version_id)

    r = api_get(url)
    return r


@model.command(name='delete-version', help='Delete version')
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', required=True)
@click.option('-v', '--version_id', 'version_id', type=str, help='Version identifier', required=True)
def delete_version(model_id, version_id):
    try:
        r = _delete_version(model_id, version_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_version(model_id, version_id):
    url = '{}/models/{}/versions/{}'.format(ORGANIZATION_ENDPOINT, model_id, version_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# deployments
# ---------------------------------------------------
@model.command(name='create-deployment', help='Deploy a specific model')
@click.option('-n', '--name', 'name', type=str, help='Deployment name', required=True)
@click.option('-m', '--model_id', 'model_id', type=str, help='Model identifier', required=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Default environment variable', default=None,
              required=False, multiple=True)
def create_deployment(name, model_id, environment):
    try:
        r = _create_deployment(name, model_id, environment)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _create_deployment(name, model_id, environment=None):
    parameter = {'name': name}
    if environment:
        parameter.update({'default_environment': dict(environment)})

    json_data = json.dumps(parameter)

    url = "{}/models/{}/deployments".format(ORGANIZATION_ENDPOINT, model_id)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-deployments', help='Get to deployment or deployment list')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', default='all',
              required=False)
def describe_deployments(deployment_id):
    try:
        r = _describe_deployments(deployment_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_deployments(deployment_id):
    url = '{}/deployments'.format(ORGANIZATION_ENDPOINT)
    if deployment_id != "all":
        url = '{}/{}'.format(url, deployment_id)
    r = api_get(url)

    return r


@model.command(name='delete-deployment', help='Delete deployment')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
def delete_deployment(deployment_id):
    try:
        r = _delete_deployment(deployment_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_deployment(deployment_id):
    url = '{}/deployments/{}'.format(ORGANIZATION_ENDPOINT, deployment_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# service
# ---------------------------------------------------
@model.command(name='create-service', help='Create http service')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-v', '--version_id', 'version_id', type=str, help='Version identifier', required=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Service environment variable', default=None,
              required=False, multiple=True)
def create_service(deployment_id, version_id, environment):

    try:
        r = _create_service(deployment_id, version_id, environment)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _create_service(deployment_id, version_id, environment=None):
    parameter = {
        'version_id': version_id
    }
    if environment:
        parameter.update({'environment': dict(environment)})
    json_data = json.dumps(parameter)
    url = "{}/deployments/{}/services".format(ORGANIZATION_ENDPOINT, deployment_id)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-services', help='Get to service or service list')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', default='all', required=False)
def describe_services(deployment_id, service_id):
    try:
        r = _describe_services(deployment_id, service_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_services(deployment_id, service_id):
    url = '{}/deployments/{}/services'.format(ORGANIZATION_ENDPOINT, deployment_id)
    if service_id != "all":
        url = '{}/{}'.format(url, service_id)
    r = api_get(url)

    return r


@model.command(name='delete-service', help='Delete service')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=True)
def delete_service(deployment_id, service_id):
    try:
        r = _delete_service(deployment_id, service_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_service(deployment_id, service_id):
    url = '{}/deployments/{}/services/{}'.format(ORGANIZATION_ENDPOINT, deployment_id, service_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# endpoint
# ---------------------------------------------------
@model.command(name='create-endpoint', help='Create endpoint')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=True)
@click.option('-c', '--custom_alias', 'custom_alias', type=str, help='Custom_alias name', required=True)
def create_endpoint(deployment_id, service_id, custom_alias):
    try:
        r = _create_endpoint(deployment_id, service_id, custom_alias)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _create_endpoint(deployment_id, service_id, custom_alias):
    parameter = {
        'service_id': service_id,
        'custom_alias': custom_alias,
    }
    json_data = json.dumps(parameter)
    time.sleep(15)
    url = "{}/deployments/{}/endpoints".format(ORGANIZATION_ENDPOINT, deployment_id)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-endpoints', help='Get to endpoint or endpoint list')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-e', '--endpoint_id', 'endpoint_id', type=str, help='Endpoint identifier', default='all', required=False)
def describe_endpoints(deployment_id, endpoint_id):
    try:
        r = _describe_endpoints(deployment_id, endpoint_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_endpoints(deployment_id, endpoint_id):
    url = '{}/deployments/{}/endpoints'.format(ORGANIZATION_ENDPOINT, deployment_id)
    if endpoint_id != "all":
        url = '{}/{}'.format(url, endpoint_id)
    r = api_get(url)

    return r


@model.command(name='delete-endpoint', help='Delete endpoint')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-e', '--endpoint_id', 'endpoint_id', type=str, help='Endpoint identifier', required=True)
def delete_endpoint(deployment_id, endpoint_id):
    try:
        r = _delete_endpoint(deployment_id, endpoint_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_endpoint(deployment_id, endpoint_id):
    url = '{}/deployments/{}/endpoints/{}'.format(ORGANIZATION_ENDPOINT, deployment_id, endpoint_id)
    r = api_delete(url)

    return r


@model.command(name='update-endpoint', help='Update endpoint')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=True)
@click.option('-e', '--endpoint_id', 'endpoint_id', type=str, help='Endpoint identifier', required=True)
def update_endpoint(deployment_id, service_id, endpoint_id):
    try:
        r = _update_endpoint(deployment_id, service_id, endpoint_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _update_endpoint(deployment_id, service_id, endpoint_id):
    parameter = {
        'service_id': service_id,
    }
    jsonData = json.dumps(parameter)
    time.sleep(15)
    url = "{}/deployments/{}/endpoints/{}".format(ORGANIZATION_ENDPOINT, deployment_id, endpoint_id)
    r = api_patch(url, jsonData)

    return r


@model.command(name='check-endpoint-json', help='Check endpoint json')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=False)
@click.option('-k', '--key', 'key', type=str, help='Key', required=True)
@click.option('-v', '--value', 'value', type=str, help='Value', required=True)
def check_endpoint_json(deployment_id, service_id, key, value):
    try:
        r = _check_endpoint_json(deployment_id, service_id, key, value)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _check_endpoint_json(deployment_id, service_id, key, value):
    parameter = {
        key: value,
    }
    json_data = json.dumps(parameter)

    if service_id:
        url = "{}/deployments/{}/services/{}".format(WEB_API_ENDPOINT, deployment_id, service_id)
    else:
        url = "{}/deployments/{}".format(WEB_API_ENDPOINT, deployment_id)
    r = api_post(url, json_data)

    return r


@model.command(name='check-endpoint-image', help='Check endpoint image')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=False)
@click.option('-t', '--type', 'type', type=str, help='Contant type', required=True)
@click.option('-i', '--image_path', 'image_path', type=str, help='Image file', required=True)
def check_endpoint_image(deployment_id, service_id, type, image_path):
    try:
        r = _check_endpoint_image(deployment_id, service_id, type, image_path)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _check_endpoint_image(deployment_id, service_id, type, image_path):
    if not os.path.exists(image_path):
        click.echo('image {} not found'.format(image_path))
        raise click.Abort()

    config = _json_file_load(CONFIG_FILE_PATH)
    if service_id:
        url = "{}/deployments/{}/services/{}".format(WEB_API_ENDPOINT, deployment_id, service_id)
    else:
        url = "{}/deployments/{}".format(WEB_API_ENDPOINT, deployment_id)
    content_type = "image/{}".format(type)
    with open(image_path, 'rb') as f:
        data = f.read()

    r = requests.post(url, data=data, headers={'Content-Type': content_type},
                      auth=(config['abeja-platform-user'], config['personal-access-token']))

    return r.json()


# ---------------------------------------------------
# submit-run
# ---------------------------------------------------
@model.command(name='submit-run', help='Submit run')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-v', '--version_id', 'version_id', type=str, help='Version identifier', required=True)
@click.option('-i', '--input_operator', 'input_operator', type=str, required=True, help='Input oparator.')
@click.option('-i', '--input_target', 'input_target', type=str, required=True, help='Input target.')
@click.option('-o', '--output_operator', 'output_operator', type=str, help='Output operator', required=False)
@click.option('-i', '--output_target', 'output_target', type=str, help='Output target', required=False)
@click.option('--retry_count', 'retry_count', type=int, required=False, default=RUN_DEFAULT_RETRY_COUNT,
              help='Retry count. By default, retries {} times.'.format(RUN_DEFAULT_RETRY_COUNT))
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Trigger environment variable', default=None,
              required=False, multiple=True)
def submit_run(deployment_id, version_id, input_operator, input_target,
               output_operator, output_target, retry_count, environment):
    if output_operator is None and output_target:
        raise click.BadArgumentUsage('Error: Missing option "--output_operator", '
                                     'when option "--output_target" is given.')
    elif output_operator and output_target is None:
        raise click.BadArgumentUsage('Error: Missing option "--output_target", '
                                     'when option "--output_operator" is given.')

    try:
        user_env_vars = dict(environment)
        r = _submit_run(deployment_id, version_id, input_operator, input_target,
                        output_operator, output_target, retry_count, user_env_vars)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _submit_run(deployment_id, version_id, input_operator, input_target,
                output_operator, output_target, retry_count, environment):
    parameter = {
        'version_id': version_id,
        'input_data': {input_operator: input_target},
        'retry_count': retry_count,
        'environment': environment
    }
    if output_operator and output_target:
        parameter['output_template'] = {output_operator: output_target}

    json_data = json.dumps(parameter)
    url = "{}/deployments/{}/runs".format(ORGANIZATION_ENDPOINT, deployment_id)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-runs', help='Get to run or run list')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-r', '--run_id', 'run_id', type=str, help='Run identifier', default='all', required=False)
def describe_runs(deployment_id, run_id):
    try:
        r = _describe_runs(deployment_id, run_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_runs(deployment_id, run_id):
    url = '{}/deployments/{}/runs'.format(ORGANIZATION_ENDPOINT, deployment_id)
    if run_id != "all":
        url = '{}/{}'.format(url, run_id)
    r = api_get(url)

    return r


@model.command(name='delete-run', help='Delete run')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-r', '--run_id', 'run_id', type=str, help='Run identifier', required=True)
def delete_run(deployment_id, run_id):
    try:
        r = delete_run(deployment_id, run_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_run(deployment_id, run_id):
    url = '{}/deployments/{}/runs/{}'.format(ORGANIZATION_ENDPOINT, deployment_id, run_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# trigger
# ---------------------------------------------------
@model.command(name='create-trigger', help='Create trigger')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-v', '--version_id', 'version_id', type=str, help='Version identifier', required=True)
@click.option('--input_service_name', 'input_service_name', type=str, help='Input service name', required=True)
@click.option('--input_service_id', 'input_service_id', type=str, help='Input service id', required=True)
@click.option('--output_service_name', 'output_service_name', type=str, required=False, default=None,
              help='Output service name')
@click.option('--output_service_id', 'output_service_id', type=str, required=False, default=None,
              help='Output service id')
@click.option('--retry_count', 'retry_count', type=int, required=False, default=TRIGGER_DEFAULT_RETRY_COUNT,
              help='Retry count. By default, retries {} times.'.format(TRIGGER_DEFAULT_RETRY_COUNT))
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Trigger environment variable', default=None,
              required=False, multiple=True)
def create_trigger(deployment_id, version_id, input_service_name, input_service_id, output_service_name,
                   output_service_id, retry_count, environment):
    if output_service_name is None and output_service_id:
        raise click.BadArgumentUsage('Error: Missing option "--output_service_name", '
                                     'when option "--output_service_id" is given.')
    elif output_service_name and output_service_id is None:
        raise click.BadArgumentUsage('Error: Missing option "--output_service_id", '
                                     'when option "--output_service_name" is given.')

    try:
        user_env_vars = dict(environment)
        r = _create_trigger(deployment_id, version_id, input_service_name, input_service_id, output_service_name,
                            output_service_id, retry_count, user_env_vars)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _create_trigger(deployment_id, version_id, input_service_name, input_service_id, output_service_name,
                    output_service_id, retry_count, environment):
    parameter = {
        'version_id': version_id,
        'input_service_name': input_service_name,
        'input_service_id': input_service_id,
        'retry_count': retry_count,
        'environment': environment
    }
    if output_service_name:
        parameter['output_service_name'] = output_service_name

    if output_service_id:
        parameter['output_service_id'] = output_service_id

    json_data = json.dumps(parameter)

    url = "{}/deployments/{}/triggers".format(ORGANIZATION_ENDPOINT, deployment_id)
    r = api_post(url, json_data)

    return r


@model.command(name='describe-triggers', help='Get to trigger or trigger list')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-t', '--trigger_id', 'trigger_id', type=str, help='Trigger identifier', default='all', required=False)
def describe_triggers(deployment_id, trigger_id):
    try:
        r = _describe_triggers(deployment_id, trigger_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _describe_triggers(deployment_id, trigger_id):
    url = '{}/deployments/{}/triggers'.format(ORGANIZATION_ENDPOINT, deployment_id)
    if trigger_id != "all":
        url = '{}/{}'.format(url, trigger_id)
    r = api_get(url)

    return r


@model.command(name='delete-trigger', help='Delete trigger')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-t', '--trigger_id', 'trigger_id', type=str, help='Trigger identifier', required=True)
def delete_trigger(deployment_id, trigger_id):
    try:
        r = _delete_trigger(deployment_id, trigger_id)
    except:
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_trigger(deployment_id, trigger_id):
    url = '{}/deployments/{}/triggers/{}'.format(ORGANIZATION_ENDPOINT, deployment_id, trigger_id)
    r = api_delete(url)

    return r


# ---------------------------------------------------
# log
# ---------------------------------------------------
@model.command(name='describe-service-logs', help='Get service log')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-s', '--service_id', 'service_id', type=str, help='Service identifier', required=True)
@click.option('--start_time', 'start_time', type=str, help='Start time / 2017-01-01T00:00:00Z', default=None,
              required=False)
@click.option('--end_time', 'end_time', type=str, help='End time / 2017-12-31T00:00:00Z', default=None,
              required=False)
@click.option('--next_token', 'next_token', type=str, help='Next token', default=None, required=False)
def describe_service_logs(deployment_id, service_id, start_time, end_time, next_token):
    params = {
        'next_token': next_token,
        'start_time': start_time,
        'end_time': end_time
    }
    url = "{}/deployments/{}/services/{}/logs".format(ORGANIZATION_ENDPOINT, deployment_id, service_id)
    try:
        r = api_get_data(url, params)
    except:
        sys.exit(ERROR_EXITCODE)

    return click.echo(json_output_formatter(r))


@model.command(name='describe-run-logs', help='Get run log')
@click.option('-d', '--deployment_id', 'deployment_id', type=str, help='Deployment identifier', required=True)
@click.option('-r', '--run_id', 'run_id', type=str, help='Run identifier', required=True)
@click.option('--start_time', 'start_time', type=str, help='Start time', default=None, required=False)
@click.option('--end_time', 'end_time', type=str, help='End time', default=None, required=False)
@click.option('--next_token', 'next_token', type=str, help='Next token', default=None, required=False)
def describe_run_logs(deployment_id, run_id, start_time, end_time, next_token):
    params = {
        'next_token': next_token,
        'start_time': start_time,
        'end_time': end_time
    }
    url = "{}/deployments/{}/runs/{}/logs".format(ORGANIZATION_ENDPOINT, deployment_id, run_id)
    try:
        r = api_get_data(url, params)
    except:
        sys.exit(ERROR_EXITCODE)

    return click.echo(json_output_formatter(r))


@model.command(name='run-local', help='Local run commands')
@click.option('-h', '--handler', 'handler', type=str, help='Model hanlder', required=True)
@click.option('-i', '--image', 'image', type=str, help='Base-image name', required=True)
@click.option('--input', 'input', type=str, help='Input data', required=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Environment variables', default=None,
              required=False, multiple=True)
@click.option('--no_cache', is_flag=True, type=bool, help='Not use built cache', required=False)
@click.option('-q', '--quiet', is_flag=True, type=bool, help='Suppress info logs', required=False)
def run_local(handler, image, input, environment, no_cache, quiet):
    image = str(image)
    local_model = LocalModelHandler()

    if not local_model.check_installation():
        click.secho("[error] docker command is required", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if not quiet:
        click.echo("[info] preparing image : {}".format(image))
    try:
        image, tag = image.split(':')
    except ValueError:
        click.secho("[error] invalid image and tag format : {}".format(image), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if not quiet:
        click.echo("[info] building image")
    try:
        local_model.build_run_image(
            image, tag, handler, env_vars=dict(environment),
            no_cache=no_cache, stdout=click.echo)
    except:
        click.secho("[error] failed to build image", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if not quiet:
        click.echo("[info] setting up local server")
    try:
        local_server = local_model.create_local_server()
    except:
        click.secho("[error] failed to create local server", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    endpoint = local_server.endpoint
    health_check_url = '{}/health_check'.format(endpoint)
    headers, data = _get_headers_and_data(input)

    with ThreadPoolExecutor() as executor:
        def server_logging():
            for log in local_server.logs():
                click.echo(log.decode('utf-8').rstrip('\n'))

        if not quiet:
            # This thread finish when local server container is skilled,
            # because it permanently tails all coming logs from the container.
            executor.submit(server_logging)

        with LocalServerManager(local_server) as manager:
            if not quiet:
                click.echo("[info] waiting server running")
            try:
                manager.wait_until_running(health_check_url)
            except Exception as e:
                click.secho("[error] failed to run local server : {}".format(e), err=True, fg='red')
                click.secho("\n ------ Local Server Error ------ ", err=True, fg='red')
                click.secho(manager.dump_logs(), err=True, fg='red')
                sys.exit(ERROR_EXITCODE)

            if not quiet:
                click.echo("[info] sending request to model")
            # TODO: need to support get and other methods
            # post request to local server
            try:
                res = manager.send_request('post', endpoint, headers, data)
            except Exception as e:
                click.secho("[error] failed to send request : {}".format(e), err=True, fg='red')
                click.secho("\n ------ Local Server Error ------ ", err=True, fg='red')
                click.secho(manager.dump_logs(), err=True, fg='red')
                sys.exit(ERROR_EXITCODE)

    if not quiet:
        click.echo("[info] finish requesting to model")
    try:
        click.echo(json.dumps(res.json(), indent=4))
    except:
        click.echo(res.content)


def _get_headers_and_data(input):
    mimetype, _ = mimetypes.guess_type(input)
    if not mimetypes:
        click.secho("[error] not supported file format {}: ".format(input), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if not os.path.exists(input):
        click.secho("[error] specifield file doesn't exist {}: ".format(input), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if mimetype in ('application/json',):
        with open(input, 'r') as f:
            _data = f.read()
            data = json.dumps(json.loads(_data))
    elif mimetype in ('image/jpeg', 'image/png', 'image/gif'):
        with open(input, 'rb') as f:
            data = f.read()
    else:
        click.secho("[error] not supported {}: ".format(input), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)
    headers = {'Content-Type': mimetype}
    return headers, data


@model.command(name='run-local-server', help='Local run commands')
@click.option('-h', '--handler', 'handler', type=str, help='Model hanlder', required=True)
@click.option('-i', '--image', 'image', type=str, help='Base-image name', required=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Environment variables', default=None,
              required=False, multiple=True)
@click.option('--no_cache', is_flag=True, type=bool, help='Not use built cache', required=False)
def run_local_server(handler, image, environment, no_cache):
    image = str(image)
    local_model = LocalModelHandler()

    if not local_model.check_installation():
        click.secho("[error] docker command is required", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    click.echo("[info] preparing image : {}".format(image))
    try:
        image, tag = image.split(':')
    except ValueError:
        click.secho("[error] invalid image and tag format : {}".format(image), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    click.echo("[info] building image")
    try:
        local_model.build_run_image(
            image, tag, handler, env_vars=dict(environment),
            no_cache=no_cache, stdout=click.echo)
    except:
        click.secho("[error] failed to build image", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    click.echo("[info] setting up local server")
    try:
        local_server = local_model.create_local_server()
    except:
        click.secho("[error] failed to create local server", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    endpoint = local_server.endpoint
    health_check_url = '{}/health_check'.format(endpoint)

    with LocalServerManager(local_server) as manager:
        # here multithreading is used to keep emitting container logs
        # until local server gets ready
        #
        # `with` clause is not used intentionally.
        # because when exiting `with` clause, executor will wait submitted function finished.
        # but here server_logging function never ends and keeps waiting logs from container.
        executor = ThreadPoolExecutor()

        def server_logging():
            for log in local_server.logs():
                click.echo(log.decode('utf-8').rstrip('\n'))

        try:
            # This thread finish when local server container is killed,
            # because it permanently tails all coming logs from the container.
            executor.submit(server_logging)

            click.echo("[info] waiting server running")

            manager.wait_until_running(health_check_url)
            executor.shutdown(wait=False)
        except Exception as e:
            click.secho("[error] failed to run local server : {}".format(e), err=True, fg='red')
            click.secho("\n ------ Local Server Error ------ ", err=True, fg='red')
            click.secho(manager.dump_logs(), err=True, fg='red')
            executor.shutdown(wait=False)
            sys.exit(ERROR_EXITCODE)

        click.secho(" ----- Local Server ----- ", fg='green')
        click.secho(" Started successfully!\n", fg='green', bold=True)

        click.secho(" Endpoint : {}".format(endpoint), bold=True)
        click.secho(" Handler :  {}".format(handler), bold=True)
        click.secho(" Image :    {}:{}".format(image, tag), bold=True)

        click.secho("\n you can now access this http api!")

        click.secho("\n press Ctrl + C to stop", bold=True)
        click.secho(" ------------------------ ", fg='green')

        try:
            # when Ctrl + C pressed, KeyboardInterrupt exception raises here.
            # then LocalServerManager#__exit__ will stop local_server.
            #
            # `since` is passed to skip logs which are already displayed until now.
            for log in local_server.logs(since=datetime.datetime.now()):
                click.echo(log.decode('utf-8').rstrip('\n'))
        except Exception as e:
            click.secho("[error] failed to send request : {}".format(e), err=True, fg='red')
            click.secho("\n ------ Local Server Error ------ ", err=True, fg='red')
            click.secho(manager.dump_logs(), err=True, fg='red')
            sys.exit(ERROR_EXITCODE)


# ---------------------------------------------------
# datalake command
# ---------------------------------------------------
@datalake.command(name='create-channel', help='Create DataLake channel')
@click.option('-n', '--name', type=str, help='Name', required=False)
@click.option('-d', '--description', type=str, help='Description', required=False)
def create_datalake_channel(name, description):
    try:
        parameters = {
            'name': name,
            'description': description,
            'storage_type': 'datalake'
        }
        parameters = {k: v for k, v in parameters.items() if v is not None}

        json_data = json.dumps(parameters)
        url = "{}/channels".format(ORGANIZATION_ENDPOINT)
        r = api_post(url, json_data)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@datalake.command(name='describe-channels', help='Describe DataLake channels')
@click.option('-c', '--channel_id', type=str, help='Channel id', default='all', required=False)
def describe_datalake_channels(channel_id):
    try:
        r = _describe_channels(channel_id, 'datalake')
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@datalake.command(name='upload', help='Upload file or directory')
@click.argument('paths', nargs=-1, type=click.Path(exists=True, resolve_path=True))
@click.option('-c', '--channel_id', 'channel_id', type=str, help='Channel identifier', required=True)
@click.option('--dry-run', '--dry_run', 'dry_run', is_flag=True, help='Dry run, only shows upload candidate files')
@click.option('-r', '--recursive', 'recursive', is_flag=True, help='Recursively upload directory')
@click.option('-m', '--metadata', '--meta-data', '--meta_data', 'metadata',
              type=METADATA_STR, multiple=True,
              help='Metadata to add all upload files')
@click.option('-l', '--file-list', '--file_list', 'file_list_path',
              type=click.Path(exists=True, resolve_path=True),
              help='JSON file which list files and metadata')
@click.pass_context
def file_upload(ctx, paths, channel_id, recursive, dry_run, metadata, file_list_path=None):
    if len(paths) == 0 and not file_list_path:
        click.secho("[error] No file specified", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    try:
        upload_file_iter = generate_upload_file_iter(
            paths=paths,
            recursive=recursive,
            file_list_path=file_list_path)
    except InvalidPathException as e:
        click.secho("[error] invalid path {}: ".format(e.path), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if dry_run:
        paths = []
        for upload_file in upload_file_iter:
            paths.append(upload_file.path)

        click.echo("[info] upload files:")
        click.echo('    ' + '\n    '.join(sorted(paths)))
        sys.exit(SUCCESS_EXITCODE)

    upload_to_datalake(channel_id, upload_file_iter, {'metadata': metadata})


@datalake.command(name='download', help='Download files')
@click.option('-c', '--channel_id', 'channel_id', type=str, help='Channel identifier', required=True)
@click.option('-o', '--output_path', 'output_path', type=str, help='Output directory path', required=True)
@click.option('-f', '--file_id', 'file_id', type=str, help='File identifier', multiple=True,
              cls=MutuallyExclusiveAndRequireOption, mutually_exclusive=["start", "end"])
@click.option('-s', '--start', 'start', type=DATE_STR, help='Start date',
              cls=MutuallyExclusiveAndRequireOption, mutually_exclusive=["file_id"], requires=['end'])
@click.option('-e', '--end', 'end', type=DATE_STR, help='End date',
              cls=MutuallyExclusiveAndRequireOption, mutually_exclusive=["file_id"], requires=['start'])
@click.option('--dry-run', '--dry_run', 'dry_run', is_flag=True, help='Dry run, only shows upload candidate files')
@click.pass_context
def file_download(ctx, channel_id, output_path, file_id, start, end, dry_run):
    if file_id:
        file_iter = generate_channel_file_iter_by_id(channel_id, *file_id)
    elif start and end:
        file_iter = generate_channel_file_iter_by_period(channel_id, start=start, end=end)
    else:
        file_iter = generate_channel_file_iter_by_period(channel_id)

    if not os.path.isdir(output_path):
        click.echo("[error] output path {} is not a directory".format(output_path), err=True)
        sys.exit(ERROR_EXITCODE)
    if dry_run:
        click.echo("[info] download files:")
        show_download_files(file_iter)
        sys.exit(SUCCESS_EXITCODE)

    download_from_datalake(channel_id, file_iter, output_path)


def show_download_files(file_iter):
    def format_file_info(file_info):
        file_id = file_info.get('file_id')
        file_meta = file_info.get('metadata', {})
        file_name = file_meta.get('x-abeja-meta-filename')
        if file_name:
            return '{}: {}'.format(file_id, file_name)
        return file_id
    files = list(map(format_file_info, file_iter))
    click.echo('    ' + '\n    '.join(files))


# ---------------------------------------------------
# schema command
# ---------------------------------------------------

def validate_content_type(ctx, param, value):
    content_type = value
    content_types = [
        'text/csv',
        'application/json',
        'application/x-jsonlines'
    ]
    if content_type not in content_types:
        raise click.BadParameter('supported content_types: (text/csv, application/json, application/x-jsonlines)')
    return value


@schema.command(name='create-schema', help='Create schema')
@click.option('-n', '--name', type=str, help='Schema name', required=True)
@click.option('-s', '--schema', type=click.File('r'), help='Schema file path', required=True)
def create_schema(name, schema):
    try:
        schema = schema.read()
        r = _create_schema(name, schema)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _create_schema(name, schema):
    parameters = {
        'display_name': name,
        'schema': json.loads(schema)
    }

    json_data = json.dumps(parameters)

    url = "{}/schemas".format(ORGANIZATION_ENDPOINT)

    r = api_post(url, json_data)
    return r


@schema.command(name='describe-schemas', help='Get schema details')
@click.option('-s', '--schema_id', type=str, help='Schema id', default='all', required=False)
def describe_schema(schema_id):
    try:
        r = _describe_schemas(schema_id)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _describe_schemas(schema_id):
    url = "{}/schemas".format(ORGANIZATION_ENDPOINT)
    if schema_id != "all":
        url = "{}/{}".format(url, schema_id)

    r = api_get(url)
    return r


# ---------------------------------------------------
# database command
# ---------------------------------------------------
@database.command(name='create-channel', help='Create database channel')
@click.option('-n', '--name', type=str, help='Name', required=True)
@click.option('-d', '--description', type=str, help='Description', required=True)
@click.option('-s', '--schema_id', type=str, help='Schema id', required=True)
def create_database_channel(name, description, schema_id):
    try:
        r = _create_database_channel(name, description, schema_id)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _create_database_channel(name, description, schema_id):
    parameters = {
        'name': name,
        'description': description,
        'schema_id': schema_id,
        'storage_type': 'rdb'
    }

    json_data = json.dumps(parameters)
    url = "{}/datamarts".format(ORGANIZATION_ENDPOINT)
    r = api_post(url, json_data)
    return r


@database.command(name='describe-channels', help='Describe database channels')
@click.option('-c', '--channel_id', type=str, help='Channel id', default='all', required=False)
def describe_database_channels(channel_id):
    try:
        r = _describe_channels(channel_id, 'rdb')
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _describe_channels(channel_id, storage_type):
    channel_limit = 1000
    if channel_id == 'all':
        url = "{}/channels?limit={}".format(ORGANIZATION_ENDPOINT, channel_limit)
        r = api_get(url)
        filtered_channels = list(filter(lambda x: x['storage_type'] == storage_type, r['channels']))
        response = {
            "created_at": r["created_at"],
            "organization_name": r["created_at"],
            "organization_id": r["organization_id"],
            "channels": filtered_channels,
            "updated_at": r['updated_at']
        }
    else:
        url = "{}/channels/{}".format(ORGANIZATION_ENDPOINT, channel_id)
        response = api_get(url)
        if response['channel']['storage_type'] != storage_type:
            response = {}
    return response


@database.command(name='describe-channel-schema', help='Get schema for channel')
@click.option('-c', '--channel_id', type=str, help='Database channel id', required=True)
def describe_channel_schema(channel_id):
    try:
        url = "{}/datamarts/{}/schema".format(ABEJA_API_URL, channel_id)
        r = api_get(url)
    except:
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


# ---------------------------------------------------
# training command
# ---------------------------------------------------
@training.command(name='init', help='Initialize training with training definition name')
@click.argument('name', required=True)
def initialize_training(name):
    try:
        training_config.write(name)
    except Exception as e:
        logger.error('training initialization failed: {}'.format(e))
        click.echo('training initialization failed.')
        sys.exit(ERROR_EXITCODE)
    click.echo('training initialized')


@training.command(name='create-job-definition', help='Create training job definition')
def create_job_definition():
    try:

        config_data = training_config.read(training_config.default_schema)
        params = {
            "name": config_data['name']
        }
        url = "{}/training/definitions".format(ORGANIZATION_ENDPOINT)
        json_data = json.dumps(params)
        r = api_post(url, json_data)
    except ConfigFileNotFoundError:
        click.echo('training configuration file does not exists.')
        logger.error('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create job definition aborted:{}'.format(e))
        click.echo('create job definition aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='create-version')
def create_training_version():
    try:
        params = training_config.read(training_config.create_version_schema)

        archive = _version_archive(params['name'])
        with open(archive.name, 'rb') as archive_file:
            source = archive_file.read()
        params["source_code_base64"] = base64.b64encode(source).decode('utf-8')
        url = "{}/training/definitions/{}/versions".format(ORGANIZATION_ENDPOINT, params['name'])
        if params.get('params'):
            params['user_parameters'] = params['params']
        json_data = json.dumps(params)
        r = api_post(url, json_data)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except FileNotFoundError:
        logger.error('failed to archive the current directory.')
        click.echo('failed to archive the current directory.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training version aborted: {}'.format(e))
        click.echo('create training version aborted.')
        sys.exit(ERROR_EXITCODE)
    finally:
        if os.path.exists(archive.name):
            archive.close()
            os.unlink(archive.name)
    click.echo(json_output_formatter(r))


@training.command(name='create-job', help='Create training job')
@click.option('-v', '--version', type=str, help='Job definition version', required=True)
@click.option('-p', '--params', type=USER_PARAM_STR, help='User parameters', default=None, required=False,
              multiple=True)
def create_training_job(version, params):
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/versions/{}/jobs".format(ORGANIZATION_ENDPOINT, config_data['name'], version)
        user_params = dict((x, y) for x, y in params)
        json_data = json.dumps({'user_parameters': user_params})
        r = api_post(url, json_data)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('create training job aborted: {}'.format(e))
        click.echo('create training job aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='describe-jobs', help='Show training job list for a specific training definition name')
def describe_jobs():
    try:
        config_data = training_config.read(training_config.default_schema)
        url = "{}/training/definitions/{}/jobs".format(ORGANIZATION_ENDPOINT, config_data['name'])
        r = api_get(url)
    except ConfigFileNotFoundError:
        logger.error('training configuration file does not exists.')
        click.echo('training configuration file does not exists.')
        sys.exit(ERROR_EXITCODE)
    except InvalidConfigException as e:
        logger.error('invalid training configuration file: {}'.format(e))
        click.echo('invalid training configuration file.')
        sys.exit(ERROR_EXITCODE)
    except Exception as e:
        logger.error('describe job operation aborted: {}'.format(e))
        click.echo('describe job operation aborted.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


@training.command(name='train-local', help='Local train commands')
@click.option('-h', '--handler', 'handler', type=str, help='Model hanlder', required=True)
@click.option('-i', '--image', 'image', type=str, help='Base-image name', required=True)
@click.option('-o', '--organization_id', 'organization_id', type=str, help='Organization ID', required=True)
@click.option('-d', '--datasets', 'datasets', type=DATASET_PARAM_STR, help='Datasets name', default=None,
              multiple=True)
@click.option('-e', '--environment', type=ENVIRONMENT_STR, help='Environment variables', default=None,
              required=False, multiple=True)
@click.option('--no_cache', is_flag=True, type=bool, help='Not use built cache', required=False)
@click.option('-q', '--quiet', is_flag=True, type=bool, help='Suppress info logs', required=False)
def train_local(handler, image, organization_id, datasets, environment, no_cache, quiet):
    image = str(image)
    local_model = LocalModelHandler()

    if not local_model.check_installation():
        click.secho("[error] docker command is required", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    if not quiet:
        click.echo("[info] preparing image : {}".format(image))
    try:
        image, tag = image.split(':')
    except ValueError:
        click.secho("[error] invalid image and tag format : {}".format(image), err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    config = _json_file_load(CONFIG_FILE_PATH)

    if not quiet:
        click.echo("[info] building image")
    try:
        local_model.build_train_image(
            image, tag, handler,
            organization_id=organization_id,
            user_id=config['abeja-platform-user'],
            access_token=config['personal-access-token'],
            datasets=dict(datasets) if datasets else {},
            env_vars=dict(environment),
            no_cache=no_cache, stdout=click.echo)
    except:
        click.secho("[error] failed to build image", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    try:
        local_container = local_model.run_image()
    except:
        click.secho("[error] failed to create local server", err=True, fg='red')
        sys.exit(ERROR_EXITCODE)

    try:
        for ret in local_container.logs(stream=True):
            sys.stdout.write(ret.decode('utf-8'))
    except KeyboardInterrupt:
        click.echo("[info] stopping, please wait for a while")
        local_model.stop_image(local_container)


# ---------------------------------------------------
# dataset command
# ---------------------------------------------------
@dataset.command(name='create-dataset', help='Create dataset channel')
@click.option('-n', '--name', type=str, help='Display name', required=True)
@click.option('-t', '--type', 'dataset_type', type=str,
              help='Type of dataset ex. classification, detection', required=True)
@click.option('-p', '--props', type=click.File('r'), help='Format to validate annotation attributes.', required=True)
def create_dataset(name, dataset_type, props):
    try:
        props = props.read()
        props = json.loads(props)
        r = _create_dataset(name, dataset_type, props)
    except Exception as e:
        logger.error('create-dataset failed: {}'.format(e))
        click.echo('create-dataset failed.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _create_dataset(name, dataset_type, props):
    parameters = {
        'name': name,
        'type': dataset_type,
        'props': props
    }

    json_data = json.dumps(parameters)
    url = "{}/datasets".format(ORGANIZATION_ENDPOINT)
    r = api_post(url, json_data)
    return r


@dataset.command(name='delete-dataset', help='Delete dataset')
@click.option('-d', '--dataset_id', type=str, help='Dataset id', required=True)
def delete_dataset(dataset_id):
    try:
        r = _delete_dataset(dataset_id)
    except Exception as e:
        logger.error('delete-dataset failed: {}'.format(e))
        click.echo('delete-dataset failed.')
        sys.exit(ERROR_EXITCODE)

    click.echo(json_output_formatter(r))


def _delete_dataset(dataset_id):
    url = '{}/datasets/{}'.format(ORGANIZATION_ENDPOINT, dataset_id)
    r = api_delete(url)

    return r


@dataset.command(name='describe-datasets', help='Describe dataset')
@click.option('-d', '--dataset_id', 'dataset_id', type=str,
              help='Dataset id', default='all', required=False)
def describe_datasets(dataset_id):
    try:
        r = _describe_datasets(dataset_id)
    except Exception as e:
        logger.error('describe-dataset failed: {}'.format(e))
        click.echo('describe-dataset failed.')
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


def _describe_datasets(dataset_id):
    url = "{}/datasets".format(ORGANIZATION_ENDPOINT) if dataset_id == 'all' \
          else "{}/datasets/{}".format(ORGANIZATION_ENDPOINT, dataset_id)
    r = api_get(url)
    return r


@dataset.command(name='import-from-datalake')
@click.option('-c', '--channel_id', 'channel_id', type=str, help='DataLake channel id', required=True)
@click.option('-d', '--dataset_id', 'dataset_id', type=str, help='Dataset id', required=True)
@click.option('--label-metadata-name', 'label_metadata', type=str,
              help='DataLake metadata that is used as label. x-abeja-meta-label is used as default.', default='label')
@click.option('--max-size-for-label', '_max', type=int, required=False,
              help='Max number of items for each labels that is uploaded to dataset API', default=None)
def dataset_import_from_datalake(channel_id, dataset_id, label_metadata, _max):
    """Imports dataset items from datalake.
    You can import dataset items from a datalake channel with labels.
    You have to prepare a datalake channel that has files you want to use as dataset.
    `x-abeja-meta-label` is used as label by default.
    For example, `cat00001.jpg` with `x-abeja-meta-label:cat` is registered to a dataset with label `cat`.
    You can specify the metadata name for label with `--label-metadata` option.
    """
    try:
        r = import_dataset_from_datalake(channel_id, dataset_id, label_metadata, _max)
    except Exception as e:
        logger.exception(e)
        sys.exit(ERROR_EXITCODE)
    click.echo(json_output_formatter(r))


# ---------------------------------------------------
# util
# ---------------------------------------------------
def _json_loads(r):
    try:
        r = json.loads(r)
    except JSONDecodeError as e:
        click.echo('{} {} {} {}'.format("[error]", date, "error message =>", e), err=True)

    return r


def _json_file_load(path):
    with open(path, 'r') as f:
        json_data = json.load(f)
    return json_data


if __name__ == '__main__':
    main()
